from .attachment import AttachmentOption, AttachmentParams
from .bool import BoolOption, BoolParams
from .channel import ChannelOption, ChannelParams
from .custom import (
    ColorOption,
    ColorParams,
    ColourOption,
    ColourParams,
    EmojiOption,
    EmojiParams,
    MemberOption,
    MemberParams,
)
from .float import FloatOption, FloatParams
from .int import IntOption, IntParams
from .mentionable import MentionableOption, MentionableParams
from .role import RoleOption, RoleParams
from .str import StrOption, StrParams
from .user import UserOption, UserParams

__all__ = (
    "AttachmentOption",
    "AttachmentParams",
    "BoolOption",
    "BoolParams",
    "ChannelOption",
    "ChannelParams",
    "ColorOption",
    "ColorParams",
    "ColourOption",
    "ColourParams",
    "EmojiOption",
    "EmojiParams",
    "FloatOption",
    "FloatParams",
    "IntOption",
    "IntParams",
    "MemberOption",
    "MemberParams",
    "MentionableOption",
    "MentionableParams",
    "RoleOption",
    "RoleParams",
    "StrOption",
    "StrParams",
    "UserOption",
    "UserParams",
)

# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
